import asyncio as aio
import logging
import typing as ty
import uuid
from dataclasses import dataclass
from enum import Enum

from ..protocols.soma import MotorCommandCodes, SomaProtocol
from .base import SENSOR_DOMAIN, BaseCover, ConnectionMode, CoverRunState
from .uuids import BATTERY, SOFTWARE_VERSION

_LOGGER = logging.getLogger(__name__)

BATTERY_ENTITY = 'battery'
ILLUMINANCE_ENTITY = 'illuminance'

POSITION_UUID = uuid.UUID('00001525-b87f-490c-92cb-11ba5ea5167c')
MOVE_PERCENT_UUID = uuid.UUID('00001526-b87f-490c-92cb-11ba5ea5167c')
MOTOR_UUID = uuid.UUID('00001530-b87f-490c-92cb-11ba5ea5167c')
NOTIFY_UUID = uuid.UUID('00001531-b87f-490c-92cb-11ba5ea5167c')
GROUP_UUID = uuid.UUID('00001893-b87f-490c-92cb-11ba5ea5167c')
NAME_UUID = uuid.UUID('00001892-b87f-490c-92cb-11ba5ea5167c')
CALIBRATE_UUID = uuid.UUID('00001529-b87f-490c-92cb-11ba5ea5167c')
STATE_UUID = uuid.UUID('00001894-b87f-490c-92cb-11ba5ea5167c')
CONFIG_UUID = uuid.UUID('00001896-b87f-490c-92cb-11ba5ea5167c')


class MovementType(Enum):
    STOP = 0
    POSITION = 1


@dataclass
class SomaState:
    battery: ty.Optional[int] = None
    position: int = 0
    illuminance: int = 0
    motor_speed: int = 0
    run_state: CoverRunState = CoverRunState.CLOSED
    target_position: ty.Optional[int] = None


class SomaCover(SomaProtocol, BaseCover):
    NAME = 'soma_shades'
    MANUFACTURER = 'Soma'

    NAME_CHAR = NAME_UUID
    POSITION_CHAR = POSITION_UUID
    MOTOR_CHAR = MOTOR_UUID
    SET_POSITION_CHAR = MOVE_PERCENT_UUID
    CHARGING_CHAR = STATE_UUID
    BATTERY_CHAR = BATTERY
    CONFIG_CHAR = CONFIG_UUID

    ACTIVE_SLEEP_INTERVAL = 1
    SEND_DATA_PERIOD = 5
    STANDBY_SEND_DATA_PERIOD_MULTIPLIER = 12 * 5  # 5 minutes
    ACTIVE_CONNECTION_MODE = ConnectionMode.ACTIVE_KEEP_CONNECTION

    # HA notation. We convert value on setting and receiving data
    CLOSED_POSITION = 0
    OPEN_POSITION = 100

    @property
    def entities(self):
        return {
            **super().entities,
            SENSOR_DOMAIN: [
                {
                    'name': BATTERY_ENTITY,
                    'device_class': 'battery',
                    'unit_of_measurement': '%',
                    'entity_category': 'diagnostic',
                },
                {
                    'name': ILLUMINANCE_ENTITY,
                    'device_class': 'illuminance',
                    'unit_of_measurement': 'lx',
                },
            ],
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self._model = 'AM43'
        self._state = SomaState()
        self.initial_status_sent = False

    def get_values_by_entities(self):
        return {
            BATTERY_ENTITY: self._state.battery,
            ILLUMINANCE_ENTITY: self._state.illuminance,
            self.COVER_ENTITY: {
                'state': self._state.run_state.value,
                'position': self._state.position,
            },
        }

    async def get_device_data(self):
        await super().get_device_data()
        name = await self._read_with_timeout(self.NAME_CHAR)
        if isinstance(name, (bytes, bytearray)):
            self._model = name.decode().strip(' \0')
        version = await self.client.read_gatt_char(SOFTWARE_VERSION)
        if version:
            self._version = version.decode()
        _LOGGER.debug(f'{self} name: {name}, version: {version}')

        cb = self.notification_callback
        await self.client.start_notify(self.CONFIG_CHAR, cb)
        await self.client.start_notify(self.POSITION_CHAR, cb)
        await self.client.start_notify(self.CHARGING_CHAR, cb)
        await self._update_full_state()

    def _handle_position(self, value):
        self._state.position = value

    def _handle_charging(self, *, charging_level, panel_level):
        self._state.illuminance = charging_level

    def _handle_motor_run_state(self, run_state: MotorCommandCodes):
        # Ignore run state.
        # We calculate run state on position and target position
        pass

    async def _update_full_state(self):
        await self._update_running_state()
        self._state.battery = await self._get_battery()
        self._state.target_position = await self._get_target_position()
        self._state.illuminance = \
            (await self._get_light_and_panel())['charging_level']
        self._state.run_state = self._get_run_state()

    async def _update_running_state(self):
        self._state.position = await self._get_position()
        self._state.motor_speed = await self._get_motor_speed()

    def _get_run_state(self) -> CoverRunState:
        if self._state.target_position == self._state.position == \
                self.OPEN_POSITION:
            return CoverRunState.OPEN
        elif self._state.target_position == self._state.position == \
                self.CLOSED_POSITION:
            return CoverRunState.CLOSED
        elif self._state.target_position < self._state.position:
            return CoverRunState.CLOSING
        elif self._state.target_position > self._state.position:
            return CoverRunState.OPENING
        return CoverRunState.STOPPED

    async def handle(self, publish_topic, send_config, *args, **kwargs):
        # request every SEND_DATA_PERIOD if running and
        # SEND_DATA_PERIOD * STANDBY_SEND_DATA_PERIOD_MULTIPLIER if in
        # standby mode

        timer = 0
        while True:
            await self.update_device_data(send_config)
            # if running notify every 5 seconds, 60 sec otherwise
            is_running = self._state.run_state in [
                CoverRunState.OPENING,
                CoverRunState.CLOSING,
            ]
            multiplier = (
                1 if is_running else self.STANDBY_SEND_DATA_PERIOD_MULTIPLIER
            )

            timer += self.ACTIVE_SLEEP_INTERVAL
            if not self.initial_status_sent or \
                    timer >= self.SEND_DATA_PERIOD * multiplier:
                if is_running:
                    _LOGGER.debug(f'[{self}] check for position')
                    await self._update_running_state()
                    if self._state.position == self._state.target_position:
                        if self._state.position == self.CLOSED_POSITION:
                            _LOGGER.info(
                                f'[{self}] Minimum position reached. '
                                f'Set to CLOSED',
                            )
                            self._state.run_state = CoverRunState.CLOSED
                        elif self._state.position == self.OPEN_POSITION:
                            _LOGGER.info(
                                f'[{self}] Maximum position reached. '
                                f'Set to OPEN',
                            )
                            self._state.run_state = CoverRunState.OPEN
                else:
                    _LOGGER.debug(f'[{self}] check for full state')
                    await self._update_full_state()
                await self._notify_state(publish_topic)
                timer = 0
            await aio.sleep(self.ACTIVE_SLEEP_INTERVAL)



import asyncio as aio
import logging
import struct
import time
import typing as ty
import uuid
from dataclasses import dataclass
from enum import Enum

from ..protocols.xiaomi import XiaomiCipherMixin
from ..utils import format_binary
from .base import BINARY_SENSOR_DOMAIN, SENSOR_DOMAIN, ConnectionMode, Device
from .uuids import SOFTWARE_VERSION

_LOGGER = logging.getLogger(__name__)


UUID_SERVICE_KETTLE = uuid.UUID('0000fe95-0000-1000-8000-00805f9b34fb')
UUID_SERVICE_KETTLE_DATA = uuid.UUID("01344736-0000-1000-8000-262837236156")
UUID_AUTH_INIT = uuid.UUID('00000010-0000-1000-8000-00805f9b34fb')
UUID_AUTH = uuid.UUID('00000001-0000-1000-8000-00805f9b34fb')
UUID_VER = uuid.UUID('00000004-0000-1000-8000-00805f9b34fb')
UUID_STATUS = uuid.UUID('0000aa02-0000-1000-8000-00805f9b34fb')

TEMPERATURE_ENTITY = 'temperature'
KETTLE_ENTITY = 'kettle'
HEAT_ENTITY = 'heat'
AUTH_MAGIC1 = bytes([0x90, 0xCA, 0x85, 0xDE])
AUTH_MAGIC2 = bytes([0x92, 0xAB, 0x54, 0xFA])

HANDLE_AUTH = 36
HANDLE_STATUS = 60


class Mode(Enum):
    IDLE = 0x00
    HEATING = 0x01
    COOLING = 0x02
    KEEP_WARM = 0x03


class LEDMode(Enum):
    BOIL = 0x01
    KEEP_WARM = 0x02
    NONE = 0xFF


class KeepWarmType(Enum):
    BOIL_AND_COOLDOWN = 0x00
    HEAT_TO_TEMP = 0x01


@dataclass
class MiKettleState:
    mode: Mode = Mode.IDLE
    led_mode: LEDMode = LEDMode.NONE
    temperature: int = 0
    target_temperature: int = 0
    keep_warm_type: KeepWarmType = KeepWarmType.BOIL_AND_COOLDOWN
    keep_warm_time: int = 0

    FORMAT = '<BBHBBBHBBB'

    @classmethod
    def from_bytes(cls, response):
        # 00 ff 00 00 5a 28 00 00 00 01 18 00
        (
            mode,  # 0
            led_mode,  # 1
            _,  # 2-3
            target_temp,  # 4
            current_temp,  # 5
            keep_warm_type,  # 6
            keep_warm_time,  # 7,8
            _, _, _,  # 9, 10, 11
        ) = struct.unpack(cls.FORMAT, response)
        return cls(
            mode=Mode(mode),
            led_mode=LEDMode(led_mode),
            temperature=current_temp,
            target_temperature=target_temp,
            keep_warm_type=KeepWarmType(keep_warm_type),
            keep_warm_time=keep_warm_time,  # minutes
        )

    def as_dict(self):
        return {
            'mode': self.mode.name.lower(),
            'running_mode': self.led_mode.name.lower(),
            'temperature': self.temperature,
            'target_temperature': self.target_temperature,
            'keep_warm_type': self.keep_warm_type.name.lower(),
            'keep_warm_minutes_passed': self.keep_warm_time,
        }


class XiaomiKettle(XiaomiCipherMixin, Device):
    NAME = 'mikettle'
    MAC_TYPE = 'random'
    ACTIVE_SLEEP_INTERVAL = 1
    SEND_INTERVAL = 30
    MANUFACTURER = 'Xiaomi'
    ACTIVE_CONNECTION_MODE = ConnectionMode.ACTIVE_KEEP_CONNECTION

    def __init__(self, mac, product_id=275, token=None,
                 *args, loop, **kwargs):
        super().__init__(mac, *args, loop=loop, **kwargs)
        self._product_id = product_id
        if token:
            assert isinstance(token, str) and len(token) == 24
            self._token = bytes.fromhex(token)
        else:
            self._token = self.generate_random_token()
        self.queue: aio.Queue = None
        self._state: ty.Optional[MiKettleState] = None

    @property
    def entities(self):
        return {
            SENSOR_DOMAIN: [
                {
                    'name': KETTLE_ENTITY,
                    'icon': 'kettle',
                    'json': True,
                    'main_value': TEMPERATURE_ENTITY,
                    'unit_of_measurement': '\u00b0C',
                },
            ],
            BINARY_SENSOR_DOMAIN: [
                {
                    'name': HEAT_ENTITY,
                    'device_class': 'heat',
                },
            ],
        }

    def get_values_by_entities(self):
        return {
            KETTLE_ENTITY: self._state.as_dict(),
            HEAT_ENTITY: self._state.mode in [Mode.HEATING, Mode.KEEP_WARM],
        }

    def notification_handler(self, sender: int, data: bytearray):
        _LOGGER.debug("Notification: {0}: {1}".format(
            sender,
            format_binary(data),
        ))
        if sender == HANDLE_STATUS:
            self._state = MiKettleState.from_bytes(data)
        else:
            # possible senders: HANDLE_AUTH == 36
            self.queue.put_nowait((sender, data))

    async def auth(self):
        await self.client.write_gatt_char(
            UUID_AUTH_INIT,
            AUTH_MAGIC1,
            True,
        )
        await self.client.start_notify(UUID_AUTH, self.notification_handler)
        await self.client.write_gatt_char(
            UUID_AUTH,
            self.cipher(
                self.mix_a(self.reverse_mac(self.mac), self._product_id),
                self._token,
            ),
            True,
        )
        auth_response = await aio.wait_for(self.queue.get(), timeout=10)
        _LOGGER.debug(f'{self} auth response: {auth_response}')
        await self.client.write_gatt_char(
            UUID_AUTH,
            XiaomiCipherMixin.cipher(self._token, AUTH_MAGIC2),
            True,
        )
        await self.client.read_gatt_char(UUID_VER)
        await self.client.stop_notify(UUID_AUTH)

    async def get_device_data(self):
        self.queue = aio.Queue()
        await self.auth()
        self._model = 'MiKettle'
        version = await self.client.read_gatt_char(SOFTWARE_VERSION)
        if version:
            self._version = version.decode()
        _LOGGER.debug(f'{self} version: {version}')
        await self.client.start_notify(UUID_STATUS, self.notification_handler)

    async def handle(self, publish_topic, send_config, *args, **kwargs):
        send_time = None
        prev_state = None
        while True:
            await self.update_device_data(send_config)
            new_state = prev_state
            if self._state:
                new_state = (
                    self._state.mode,
                    self._state.led_mode,
                    self._state.keep_warm_type,
                )
            if self._state and (
                not send_time or
                (time.time() - send_time) > self.SEND_INTERVAL or
                new_state != prev_state
            ):
                send_time = time.time()
                prev_state = new_state
                await self._notify_state(publish_topic)
            await aio.sleep(self.ACTIVE_SLEEP_INTERVAL)



#!/usr/bin/env python

from setuptools import find_packages, setup

from ble2mqtt.__version__ import VERSION

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='ble2mqtt',
    version=VERSION,
    description='BLE to MQTT bridge',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Ivan Belokobylskiy',
    author_email='belokobylskij@gmail.com',
    url='https://github.com/devbis/ble2mqtt/',
    entry_points={
        'console_scripts': ['ble2mqtt=ble2mqtt.__main__:main']
    },
    packages=find_packages(include=['ble2mqtt', 'ble2mqtt.*']),
    install_requires=[
        'aio-mqtt>=0.2.0',
        'bleak>=0.12.0',
    ],
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Utilities',
    ],
)



import logging
import typing as ty
from dataclasses import dataclass
from datetime import datetime, timedelta

from bleak.backends.device import BLEDevice

from .base import BINARY_SENSOR_DOMAIN, DEVICE_TRACKER_DOMAIN, Sensor

_LOGGER = logging.getLogger(__name__)


@dataclass
class SensorState:
    presence: bool = False
    last_check: ty.Optional[datetime] = None

    @property
    def device_tracker(self):
        return 'home' if self.presence else 'not_home'


class Presence(Sensor):
    NAME = 'presence'
    SENSOR_CLASS = SensorState
    SUPPORT_PASSIVE = True
    SUPPORT_ACTIVE = False
    MANUFACTURER = 'Generic'
    THRESHOLD = 300  # if no activity more than THRESHOLD, consider presence=OFF
    PASSIVE_SLEEP_INTERVAL = 1
    SEND_DATA_PERIOD = 60

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        cls = self.SENSOR_CLASS
        self._state: cls = None
        self._threshold = int(kwargs.get('threshold', self.THRESHOLD))

    @property
    def entities(self):
        return {
            BINARY_SENSOR_DOMAIN: [
                {
                    'name': 'presence',
                    'device_class': 'presence',
                },
            ],
            DEVICE_TRACKER_DOMAIN: [
                {
                    'name': 'device_tracker',
                },
            ],
        }

    def handle_advert(self, scanned_device: BLEDevice, adv_data):
        self._state = self.SENSOR_CLASS(
            presence=True,
            last_check=datetime.now(),
        )
        _LOGGER.debug(
            f'Advert received for {self}, current state: {self._state}',
        )

    async def handle_passive(self, *args, **kwargs):
        self.last_sent_value = None
        self.last_sent_time = None
        await super().handle_passive(*args, **kwargs)

    async def do_passive_loop(self, publish_topic):
        if self._state.presence and \
                self._state.last_check + \
                timedelta(seconds=self._threshold) < datetime.now():
            self._state.presence = False
        # send if changed or update value every SEND_DATA_PERIOD secs
        if self.last_sent_value is None or \
                self.last_sent_value != self._state.presence or \
                (datetime.now() - self.last_sent_time).seconds > \
                self.SEND_DATA_PERIOD:

            _LOGGER.debug(f'Try publish {self._state}')
            await self._notify_state(publish_topic)
            self.last_sent_value = self._state.presence
            self.last_sent_time = datetime.now()



import asyncio as aio
import logging
import typing as ty
import uuid
from dataclasses import dataclass

from ble2mqtt.protocols.ensto import ActiveMode, EnstoProtocol, Measurements

from ..utils import format_binary
from .base import BaseClimate, ClimateMode, ConnectionMode
from .uuids import DEVICE_NAME, SOFTWARE_VERSION

_LOGGER = logging.getLogger(__name__)

UUID_CHILD_LOCK = uuid.UUID('6e3064e2-d9a5-4ca0-9d14-017c59627330')
UUID_MEASUREMENTS = uuid.UUID('66ad3e6b-3135-4ada-bb2b-8b22916b21d4')
UUID_VACATION = uuid.UUID('6584e9c6-4784-41aa-ac09-c899191048ae')
UUID_DATE = uuid.UUID('b43f918a-b084-45c8-9b60-df648c4a4a1e')
UUID_HEATING_POWER = uuid.UUID('53b7bf87-6cf0-4790-839a-e72d3afbec44')
UUID_FACTORY_RESET = uuid.UUID('f366dddb-ebe2-43ee-83c0-472ded74c8fa')


@dataclass
class EnstoState:
    mode: ClimateMode = ClimateMode.OFF
    temperature: ty.Optional[float] = None
    target_temperature: ty.Optional[float] = None
    relay_is_on: bool = False
    target_temperature_with_offset: ty.Optional[float] = None


class EnstoThermostat(EnstoProtocol, BaseClimate):
    NAME = 'ensto_thermostat'  # EPHBEBT
    MANUFACTURER = 'Ensto'
    ACTIVE_CONNECTION_MODE = ConnectionMode.ACTIVE_KEEP_CONNECTION

    AUTH_CHAR = UUID_FACTORY_RESET
    DATE_CHAR = UUID_DATE
    VACATION_CHAR = UUID_VACATION
    MEASUREMENTS_CHAR = UUID_MEASUREMENTS
    CUSTOM_MEMORY_SLOT_CHAR = UUID_HEATING_POWER
    DEFAULT_TARGET_TEMPERATURE = 18.0
    MIN_POTENTIOMETER_VALUE = 5.0

    SEND_DATA_PERIOD = 60

    MODES = (ClimateMode.OFF, ClimateMode.HEAT)

    def __init__(self, *args, key='', **kwargs):
        super().__init__(*args, **kwargs)
        self._state = EnstoState()
        if key:
            assert len(key) == 8, f'{self}: Key must be 8 chars long'
            self._reset_id = bytes.fromhex(key)
        self.initial_status_sent = False

    async def get_device_data(self):
        await super().get_device_data()
        _LOGGER.debug(f'{self} start protocol')
        await self.protocol_start()
        name = await self._read_with_timeout(DEVICE_NAME)
        if isinstance(name, (bytes, bytearray)):
            self._model = name[1:].decode('latin1').strip(' \0') or 'Heater'
        version = await self.client.read_gatt_char(SOFTWARE_VERSION)
        if version:
            self._version = version.decode('latin1')
        _LOGGER.debug(f'{self} name: {self._model}, version: {self._version}')

        self._state.target_temperature = await self.read_target_temp()
        _LOGGER.debug(f'{self} target temp: {self._state}')
        await self.guess_potentiometer_position()
        await self._update_state()
        previously_saved_target_temp = await self.read_target_temp()
        if previously_saved_target_temp:
            self._state.target_temperature = previously_saved_target_temp

    async def guess_potentiometer_position(self):
        _LOGGER.debug(f'{self} guess_potentiometer_position')
        values = await self.read_measurements()
        cur_vacation_mode = await self.read_vacation_mode()
        self.set_current_potentiometer_value(values, cur_vacation_mode)
        _LOGGER.debug(
            f'{self} update _heater_potentiometer_temperature '
            f'{self._heater_potentiometer_temperature}',
        )

    async def _set_target_temperature(self, value):
        await self.save_target_temp(value)
        # set vacation mode and offset
        if self._state.mode != ClimateMode.OFF:
            await self.set_vacation_mode(value)
        self._state.target_temperature = value
        await self._update_state()

    async def _switch_mode(self, next_mode: ClimateMode):
        # set vacation mode and offset
        if next_mode == ClimateMode.HEAT:
            temp = await self.read_target_temp()
            if not temp:
                temp = self.DEFAULT_TARGET_TEMPERATURE
                await self.save_target_temp(temp)
        else:
            temp = 5.0
        await self.set_vacation_mode(temp, True)
        await self._update_state()

    def get_values_by_entities(self) -> ty.Dict[str, ty.Any]:
        return {
            self.CLIMATE_ENTITY: {
                'mode': self._state.mode.value,
                'temperature': self._state.temperature,
                'target_temperature': self._state.target_temperature,
            },
        }

    def set_current_potentiometer_value(self, measurements: Measurements,
                                        vacation_data: bytes):
        offset_temp = int.from_bytes(
            vacation_data[10:12],
            byteorder='little',
            signed=True,
        )/100
        vacation_enabled = vacation_data[13]
        _LOGGER.debug(
            f'{self} vacation_data: '
            f'offset_temp={offset_temp}, vacation_enabled={vacation_enabled}',
        )
        if measurements.active_mode == ActiveMode.MANUAL:
            self._heater_potentiometer_temperature = \
                measurements.target_temperature
        elif measurements.active_mode == ActiveMode.VACATION:
            self._heater_potentiometer_temperature = \
                measurements.target_temperature - offset_temp

    async def _update_state(self):
        values = await self.read_measurements()
        _LOGGER.debug(f'{self} parsed measurements: {values}')
        self._state.temperature = values.temperature
        self._state.target_temperature_with_offset = values.target_temperature
        self._state.relay_is_on = values.relay_is_on
        vacation_data = await self.read_vacation_mode()
        _LOGGER.debug(f'{self} vacation_data: {format_binary(vacation_data)}')
        self.set_current_potentiometer_value(values, vacation_data)

        if (
            values.active_mode != ActiveMode.VACATION or
            self._state.target_temperature_with_offset >
                self.MIN_POTENTIOMETER_VALUE
        ):
            self._state.mode = ClimateMode.HEAT
            self._state.target_temperature = \
                self._state.target_temperature_with_offset
        else:
            self._state.mode = ClimateMode.OFF

    async def handle(self, publish_topic, send_config, *args, **kwargs):
        timer = 0
        while True:
            await self.update_device_data(send_config)

            timer += self.ACTIVE_SLEEP_INTERVAL
            if not self.initial_status_sent or \
                    timer >= self.SEND_DATA_PERIOD:
                _LOGGER.debug(f'[{self}] check for measurements')
                await self._update_state()
                await self._notify_state(publish_topic)
                timer = 0
            await aio.sleep(self.ACTIVE_SLEEP_INTERVAL)



import abc
import asyncio as aio
import logging
import struct
import uuid

from ..devices.base import Sensor, SubscribeAndSetDataMixin

_LOGGER = logging.getLogger(__name__)


# Xiaomi Humidity/Temperature sensors

class XiaomiPoller(SubscribeAndSetDataMixin, Sensor, abc.ABC):
    DATA_CHAR: uuid.UUID = None  # type: ignore
    BATTERY_CHAR: uuid.UUID = None  # type: ignore
    MANUFACTURER = 'Xiaomi'

    def __init__(self, *args, loop, **kwargs):
        super().__init__(*args, loop=loop, **kwargs)
        self._stack = aio.LifoQueue(loop=loop)

    def process_data(self, data):
        self._loop.call_soon_threadsafe(self._stack.put_nowait, data)

    async def read_and_send_data(self, publish_topic):
        raise NotImplementedError()

    async def handle_active(self, publish_topic, send_config, *args, **kwargs):
        _LOGGER.debug(f'Wait {self} for connecting...')
        sec_to_wait_connection = 0
        while True:
            if not self.client.is_connected:
                if sec_to_wait_connection >= 30:
                    raise TimeoutError(
                        f'{self} not connected for 30 sec in handle()',
                    )
                sec_to_wait_connection += self.NOT_READY_SLEEP_INTERVAL
                await aio.sleep(self.NOT_READY_SLEEP_INTERVAL)
                continue
            try:
                _LOGGER.debug(f'{self} connected!')
                # in case of bluetooth error populating queue
                # could stop and will wait for self._stack.get() forever
                await self.update_device_data(send_config)
                await aio.wait_for(
                    self.read_and_send_data(publish_topic),
                    timeout=15,
                )
            except ValueError as e:
                _LOGGER.error(f'[{self}] Cannot read values {str(e)}')
            else:
                await self.close()
                return
            await aio.sleep(1)


# Xiaomi Kettles

class XiaomiCipherMixin:
    # Picked from the https://github.com/drndos/mikettle/
    @staticmethod
    def generate_random_token() -> bytes:
        return bytes([  # from component, maybe random is ok
            0x01, 0x5C, 0xCB, 0xA8, 0x80, 0x0A, 0xBD, 0xC1, 0x2E, 0xB8,
            0xED, 0x82,
        ])
        # return os.urandom(12)

    @staticmethod
    def reverse_mac(mac) -> bytes:
        parts = mac.split(":")
        reversed_mac = bytearray()
        length = len(parts)
        for i in range(1, length + 1):
            reversed_mac.extend(bytearray.fromhex(parts[length - i]))
        return reversed_mac

    @staticmethod
    def mix_a(mac, product_id) -> bytes:
        return bytes([
            mac[0], mac[2], mac[5], (product_id & 0xff), (product_id & 0xff),
            mac[4], mac[5], mac[1],
        ])

    @staticmethod
    def mix_b(mac, product_id) -> bytes:
        return bytes([
            mac[0], mac[2], mac[5], ((product_id >> 8) & 0xff), mac[4], mac[0],
            mac[5], (product_id & 0xff),
        ])

    @staticmethod
    def _cipher_init(key) -> bytes:
        perm = bytearray()
        for i in range(0, 256):
            perm.extend(bytes([i & 0xff]))
        keyLen = len(key)
        j = 0
        for i in range(0, 256):
            j += perm[i] + key[i % keyLen]
            j = j & 0xff
            perm[i], perm[j] = perm[j], perm[i]
        return perm

    @staticmethod
    def _cipher_crypt(input, perm) -> bytes:
        index1 = 0
        index2 = 0
        output = bytearray()
        for i in range(0, len(input)):
            index1 = index1 + 1
            index1 = index1 & 0xff
            index2 += perm[index1]
            index2 = index2 & 0xff
            perm[index1], perm[index2] = perm[index2], perm[index1]
            idx = perm[index1] + perm[index2]
            idx = idx & 0xff
            output_byte = input[i] ^ perm[idx]
            output.extend(bytes([output_byte & 0xff]))

        return output

    @classmethod
    def cipher(cls, key, input) -> bytes:
        perm = cls._cipher_init(key)
        return cls._cipher_crypt(input, perm)


# region xiaomi advert parsers from 0xfe95

# this part is partly taken from ble_monitor component for Home Assistant

# Structured objects for data conversions
TH_STRUCT = struct.Struct("<hH")
H_STRUCT = struct.Struct("<H")
T_STRUCT = struct.Struct("<h")
TTB_STRUCT = struct.Struct("<hhB")
CND_STRUCT = struct.Struct("<H")
ILL_STRUCT = struct.Struct("<I")
LIGHT_STRUCT = struct.Struct("<I")
FMDH_STRUCT = struct.Struct("<H")
M_STRUCT = struct.Struct("<L")
P_STRUCT = struct.Struct("<H")
BUTTON_STRUCT = struct.Struct("<BBB")
FLOAT_STRUCT = struct.Struct("<f")


def obj0010(xobj):
    # Toothbrush
    if xobj[0] == 0:
        return {'toothbrush': 1, 'counter': xobj[1]}
    else:
        return {'toothbrush': 0, 'score': xobj[1]}


def obj1004(xobj):
    # Temperature
    if len(xobj) == 2:
        (temp,) = T_STRUCT.unpack(xobj)
        return {"temperature": temp / 10}
    else:
        return {}


def obj1005(xobj):
    return {"switch": xobj[0], "temperature": xobj[1]}


def obj1006(xobj):
    # Humidity
    if len(xobj) == 2:
        (humi,) = H_STRUCT.unpack(xobj)
        return {"humidity": humi / 10}
    else:
        return {}


def obj1007(xobj):
    # Illuminance
    if len(xobj) == 3:
        (illum,) = ILL_STRUCT.unpack(xobj + b'\x00')
        return {"illuminance": illum, "light": 1 if illum == 100 else 0}
    else:
        return {}


def obj1008(xobj):
    # Moisture
    return {"moisture": xobj[0]}


def obj1009(xobj):
    # Conductivity
    if len(xobj) == 2:
        (cond,) = CND_STRUCT.unpack(xobj)
        return {"conductivity": cond}
    else:
        return {}


def obj1010(xobj):
    # Formaldehyde
    if len(xobj) == 2:
        (fmdh,) = FMDH_STRUCT.unpack(xobj)
        return {"formaldehyde": fmdh / 100}
    else:
        return {}


def obj1012(xobj):
    # Switch
    return {"switch": xobj[0]}


def obj1013(xobj):
    # Consumable (in percent)
    return {"consumable": xobj[0]}


def obj1014(xobj):
    # Moisture
    return {"moisture": xobj[0]}


def obj1015(xobj):
    # Smoke
    return {"smoke detector": xobj[0]}


def obj1017(xobj):
    # Motion
    if len(xobj) == 4:
        (motion,) = M_STRUCT.unpack(xobj)
        # seconds since last motion detected message
        # (not used, we use motion timer in obj000f)
        # 0 = motion detected
        return {"motion": 1 if motion == 0 else 0}
    else:
        return {}


def obj1018(xobj):
    # Light intensity
    return {"light": xobj[0]}


def obj100a(xobj):
    # Battery
    batt = xobj[0]
    volt = 2.2 + (3.1 - 2.2) * (batt / 100)
    return {"battery": batt, "voltage": volt}


def obj100d(xobj):
    # Temperature and humidity
    if len(xobj) == 4:
        (temp, humi) = TH_STRUCT.unpack(xobj)
        return {"temperature": temp / 10, "humidity": humi / 10}
    else:
        return {}


# The following data objects are device specific.
# For now only added for XMWSDJ04MMC
# https://miot-spec.org/miot-spec-v2/instances?status=all
def obj4803(xobj):
    # Battery
    batt = xobj[0]
    return {"battery": batt}


def obj4c01(xobj):
    if len(xobj) == 4:
        temp = FLOAT_STRUCT.unpack(xobj)[0]
        return {"temperature": temp}
    else:
        return {}


def obj4c08(xobj):
    if len(xobj) == 4:
        humi = FLOAT_STRUCT.unpack(xobj)[0]
        return {"humidity": humi}
    else:
        return {}


# Dataobject dictionary
# {dataObject_id: (converter}
xiaomi_dataobject_dict = {
    # 0x0003: obj0003,
    # 0x0006: obj0006,
    0x0010: obj0010,
    # 0x000B: obj000b,
    # 0x000F: obj000f,
    # 0x1001: obj1001,
    0x1004: obj1004,
    0x1005: obj1005,
    0x1006: obj1006,
    0x1007: obj1007,
    0x1008: obj1008,
    0x1009: obj1009,
    0x1010: obj1010,
    0x1012: obj1012,
    0x1013: obj1013,
    0x1014: obj1014,
    0x1015: obj1015,
    0x1017: obj1017,
    0x1018: obj1018,
    # 0x1019: obj1019,
    0x100A: obj100a,
    0x100D: obj100d,
    # 0x2000: obj2000,
    0x4803: obj4803,
    0x4c01: obj4c01,
    0x4c08: obj4c08,
}


def parse_fe95_advert(adv_data: bytes) -> dict:
    frctrl = int.from_bytes(adv_data[:2], byteorder='little')

    # frctrl_mesh = (frctrl >> 7) & 1  # mesh device
    # frctrl_version = frctrl >> 12  # version
    # frctrl_auth_mode = (frctrl >> 10) & 3
    # frctrl_solicited = (frctrl >> 9) & 1
    # frctrl_registered = (frctrl >> 8) & 1
    # frctrl_object_include = (frctrl >> 6) & 1
    frctrl_capability_include = (frctrl >> 5) & 1
    frctrl_mac_include = (frctrl >> 4) & 1  # check for MAC address in data
    # frctrl_is_encrypted = (frctrl >> 3) & 1  # check for encryption being used
    # frctrl_request_timing = frctrl & 1  # old version

    counter = 5
    if frctrl_mac_include:
        counter += 6
    # check for capability byte present
    if frctrl_capability_include:
        counter += 1
        # capability_io = adv_data[counter - 1]

    payload = adv_data[counter:]
    result = {}
    if payload:
        payload_start = 0
        payload_length = len(payload)
        while payload_length >= payload_start + 3:
            obj_typecode = \
                payload[payload_start] + (payload[payload_start + 1] << 8)
            obj_length = payload[payload_start + 2]
            next_start = payload_start + 3 + obj_length
            if payload_length < next_start:
                _LOGGER.debug(
                    "Invalid payload data length, payload: %s", payload.hex(),
                )
                break
            object = payload[payload_start + 3:next_start]
            if obj_length != 0:
                resfunc = xiaomi_dataobject_dict.get(obj_typecode, None)
                if resfunc:
                    # if hex(obj_typecode) in ["0x1001", "0xf"]:
                    #     result.update(resfunc(object, device_type))
                    # else:
                    result.update(resfunc(object))
                else:
                    # if self.report_unknown == "Xiaomi":
                    _LOGGER.info(
                        "UNKNOWN dataobject in payload! Adv: %s",
                        adv_data.hex(),
                    )
            payload_start = next_start

    return result

# endregion



import logging
import struct
import uuid
from dataclasses import dataclass

from .base import (SENSOR_DOMAIN, ConnectionMode, Sensor,
                   SubscribeAndSetDataMixin)

_LOGGER = logging.getLogger(__name__)

MAIN_DATA = uuid.UUID('70BC767E-7A1A-4304-81ED-14B9AF54F7BD')


@dataclass
class SensorState:
    battery: int
    dose: float
    dose_rate: float
    temperature: int

    @classmethod
    def from_data(cls, sensor_data):
        flags, dose, dose_rate, pulses, battery, temp = \
            struct.unpack('<BffHbb', sensor_data)
        return cls(
            dose=round(dose, 4),
            dose_rate=round(dose_rate, 4),
            battery=battery,
            temperature=temp,
        )


class AtomFast(SubscribeAndSetDataMixin, Sensor):
    NAME = 'atomfast'
    DATA_CHAR = MAIN_DATA
    SENSOR_CLASS = SensorState
    CONNECTION_FAILURES_LIMIT = 10
    MANUFACTURER = 'Atom'
    ACTIVE_CONNECTION_MODE = ConnectionMode.ACTIVE_KEEP_CONNECTION

    @property
    def entities(self):
        return {
            SENSOR_DOMAIN: [
                {
                    'name': 'temperature',
                    'device_class': 'temperature',
                    'unit_of_measurement': '\u00b0C',
                },
                {
                    'name': 'dose',
                    'unit_of_measurement': 'mSv',
                    'icon': 'atom',
                },
                {
                    'name': 'dose_rate',
                    'unit_of_measurement': 'μSv/h',
                    'icon': 'atom',
                },
                {
                    'name': 'battery',
                    'device_class': 'battery',
                    'unit_of_measurement': '%',
                    'entity_category': 'diagnostic',
                },
            ],
        }

    def filter_notifications(self, sender):
        return sender == 0x24



from .atom_fast import AtomFast  # noqa: F401
from .base import registered_device_types  # noqa: F401
from .bulb_avea import AveaBulb  # noqa: F401
from .cooker_redmond import RedmondCooker  # noqa: F401
from .cover_am43 import AM43Cover  # noqa: F401
from .cover_soma import SomaCover  # noqa: F401
from .kettle_redmond import RedmondKettle  # noqa: F401
from .kettle_xiaomi import XiaomiKettle  # noqa: F401
from .presence import Presence  # noqa: F401
from .thermostat_ensto import EnstoThermostat  # noqa: F401
from .vson_air_wp6003 import VsonWP6003  # noqa: F401
from .xiaomi_ht import XiaomiHumidityTemperatureV1  # noqa: F401
from .xiaomi_lywsd03 import XiaomiHumidityTemperatureLYWSD  # noqa: F401
from .xiaomi_lywsd03_atc import XiaomiHumidityTemperatureLYWSDATC  # noqa: F401
